@@ -5,6 +5,123 @@ use raystack_core::{Number, Qname, Ref, Symbol, TagName};
 use std::collections::HashMap;
 use std::convert::{From, TryFrom, TryInto};
 
+#[derive(Clone, Debug, PartialEq)]
+pub struct Comp {
+    pub cells: HashMap<TagName, Dict>,
+    pub body: Block,
+}
+
+impl Comp {
+    pub fn new(cells: HashMap<TagName, Dict>, body: Block) -> Self {
+        Self { cells, body }
+    }
+}
+
+impl TryFrom<&Val> for Comp {
+    type Error = ();
+
+    fn try_from(val: &Val) -> Result<Self, Self::Error> {
+        let hash_map = map_for_type(val, "compdef").map_err(|_| ())?;
+        let body = get_val(hash_map, "body")
+            .expect("comdef should contain 'body' tag");
+        let cells = get_val(hash_map, "cells")
+            .expect("compdef should contain 'cells' tag");
+
+        let cells_dict = match cells {
+            Val::Dict(cells_map) => cells_map,
+            _ => return Err(()),
+        };
+
+        let cells_dict = cells_dict
+            .iter()
+            .map(|(key, value)| {
+                comp_cell_val_to_ast_dict(value).map(|dict| (key.clone(), dict))
+            })
+            .collect::<Option<HashMap<_, _>>>()
+            .ok_or(())?;
+
+        let body_block = body
+            .try_into()
+            .expect("compdef 'body' could not be parsed as a Block");
+
+        Ok(Self::new(cells_dict, body_block))
+    }
+}
+
+/// Takes a Val which should be a comp cell Dict, and attempts to parse it
+/// into an ast::Dict.
+fn comp_cell_val_to_ast_dict(cell_val: &Val) -> Option<Dict> {
+    match cell_val {
+        Val::Dict(hash_map) => {
+            let dict_map = hash_map
+                .clone()
+                .into_iter()
+                .map(|(key, value)| {
+                    let new_value = match value.as_ref() {
+                        Val::Lit(lit) => Some(comp_cell_lit_to_dict_val(lit)),
+                        Val::Dict(_) => {
+                            comp_cell_val_to_ast_dict(value.as_ref())
+                                .map(|dict| DictVal::Expr(Expr::Dict(dict)))
+                        }
+                        Val::List(_) => {
+                            comp_cell_val_to_ast_list(value.as_ref())
+                                .map(|list| DictVal::Expr(Expr::List(list)))
+                        }
+                    };
+                    Some((key, new_value?))
+                })
+                .collect::<Option<HashMap<_, _>>>();
+            Some(Dict::new(dict_map?))
+        }
+        _ => None,
+    }
+}
+
+fn comp_cell_lit_to_dict_val(lit: &ap::Lit) -> DictVal {
+    match lit {
+        ap::Lit::Bool(bool) => DictVal::Expr(Expr::Lit(Lit::Bool(*bool))),
+        ap::Lit::Date(date) => DictVal::Expr(Expr::Lit(Lit::Date(*date))),
+        ap::Lit::DictMarker => DictVal::Marker,
+        ap::Lit::DictRemoveMarker => DictVal::RemoveMarker,
+        ap::Lit::Null => DictVal::Expr(Expr::Lit(Lit::Null)),
+        ap::Lit::Num(number) => {
+            DictVal::Expr(Expr::Lit(Lit::Num(number.clone())))
+        }
+        ap::Lit::Ref(reff) => DictVal::Expr(Expr::Lit(Lit::Ref(reff.clone()))),
+        ap::Lit::Str(string) => {
+            DictVal::Expr(Expr::Lit(Lit::Str(string.clone())))
+        }
+        ap::Lit::Symbol(symbol) => {
+            DictVal::Expr(Expr::Lit(Lit::Symbol(symbol.clone())))
+        }
+        ap::Lit::Time(time) => DictVal::Expr(Expr::Lit(Lit::Time(*time))),
+        ap::Lit::Uri(uri) => DictVal::Expr(Expr::Lit(Lit::Uri(uri.clone()))),
+        ap::Lit::YearMonth(ym) => {
+            DictVal::Expr(Expr::Lit(Lit::YearMonth(ym.into())))
+        }
+    }
+}
+
+fn comp_cell_val_to_ast_list(list: &Val) -> Option<List> {
+    match list {
+        Val::List(list) => {
+            let exprs = list.iter().map(|value| {
+            match value {
+                Val::Lit(lit) => Some(Expr::Lit(lit.try_into().expect("expected ap::Val::Lit in a comp cell list to parse into an ast::Lit"))),
+                Val::Dict(_) => {
+                    comp_cell_val_to_ast_dict(value).map(Expr::Dict)
+                },
+                Val::List(_) => {
+                    comp_cell_val_to_ast_list(value).map(Expr::List)
+                }
+            }
+        }).collect::<Option<Vec<_>>>();
+            Some(List::new(exprs?))
+        }
+        _ => None,
+    }
+}
+
 #[derive(Clone, Copy, Debug, PartialEq)]
 pub enum Associativity {
     Left,
@@ -501,6 +618,7 @@ impl BuildDotCallsChain {
     }
 }
 
+/// A single dot call which is part of a dot call chain.
 #[derive(Debug, Clone, PartialEq)]
 pub(crate) struct ChainedDotCall {
     pub(crate) func_name: FuncName,
@@ -1308,9 +1426,9 @@ impl TryFrom<&Val> for Expr {
     type Error = ();
 
     fn try_from(val: &Val) -> Result<Self, Self::Error> {
-        let lit: Option<LitInner> = val.try_into().ok();
+        let lit: Option<Lit> = val.try_into().ok();
         if let Some(lit) = lit {
-            return Ok(Expr::Lit(Lit::new(lit)));
+            return Ok(Expr::Lit(lit));
         };
 
         if let Some(tag_name) = var_val_to_tag_name(val) {
@@ -1595,28 +1713,9 @@ impl TryFrom<&ap::Val> for Param {
     }
 }
 
-#[derive(Clone, Debug, PartialEq)]
-pub struct Lit {
-    lit: LitInner,
-}
-
-impl Lit {
-    pub fn new(lit: LitInner) -> Self {
-        Self { lit }
-    }
-
-    pub fn lit(&self) -> &LitInner {
-        &self.lit
-    }
-
-    pub fn to_axon_code(&self) -> String {
-        self.lit().to_axon_code()
-    }
-}
-
 /// A Axon literal value.
 #[derive(Clone, Debug, PartialEq)]
-pub enum LitInner {
+pub enum Lit {
     Bool(bool),
     Date(NaiveDate),
     Null,
@@ -1629,14 +1728,14 @@ pub enum LitInner {
     YearMonth(YearMonth),
 }
 
-impl LitInner {
+impl Lit {
     pub fn to_axon_code(&self) -> String {
         match self {
             Self::Bool(true) => "true".to_owned(),
             Self::Bool(false) => "false".to_owned(),
             Self::Date(d) => d.format("%Y-%m-%d").to_string(),
             Self::Null => "null".to_owned(),
-            Self::Num(n) => number_to_axon_code(n),
+            Self::Num(n) => n.to_axon_code(),
             Self::Ref(r) => r.to_axon_code().to_owned(),
             Self::Str(s) => format!("{:?}", s).replace(r"\\$", r"\$"),
             Self::Symbol(s) => s.to_axon_code().to_owned(),
@@ -1647,45 +1746,38 @@ impl LitInner {
     }
 }
 
-fn number_to_axon_code(n: &Number) -> String {
-    let scalar = n.value();
-    match n.unit() {
-        Some(unit) => format!("{}{}", scalar, unit),
-        None => format!("{}", scalar),
-    }
-}
-
 /// Enumerates the types of errors that can occur when converting to a
 /// `Lit`.
+#[derive(Clone, Copy, Debug, PartialEq)]
 pub enum ConvertLitError {
     IsDictMarker,
     IsDictRemoveMarker,
 }
 
-impl TryFrom<&ap::Lit> for LitInner {
+impl TryFrom<&ap::Lit> for Lit {
     type Error = ConvertLitError;
 
     fn try_from(lit: &ap::Lit) -> Result<Self, Self::Error> {
         match lit {
-            ap::Lit::Bool(bool) => Ok(LitInner::Bool(*bool)),
-            ap::Lit::Date(date) => Ok(LitInner::Date(*date)),
+            ap::Lit::Bool(bool) => Ok(Lit::Bool(*bool)),
+            ap::Lit::Date(date) => Ok(Lit::Date(*date)),
             ap::Lit::DictMarker => Err(ConvertLitError::IsDictMarker),
             ap::Lit::DictRemoveMarker => {
                 Err(ConvertLitError::IsDictRemoveMarker)
             }
-            ap::Lit::Null => Ok(LitInner::Null),
-            ap::Lit::Num(number) => Ok(LitInner::Num(number.clone())),
-            ap::Lit::Ref(reff) => Ok(LitInner::Ref(reff.clone())),
-            ap::Lit::Str(string) => Ok(LitInner::Str(string.clone())),
-            ap::Lit::Symbol(symbol) => Ok(LitInner::Symbol(symbol.clone())),
-            ap::Lit::Time(time) => Ok(LitInner::Time(*time)),
-            ap::Lit::Uri(uri) => Ok(LitInner::Uri(uri.clone())),
-            ap::Lit::YearMonth(ym) => Ok(LitInner::YearMonth(ym.into())),
+            ap::Lit::Null => Ok(Lit::Null),
+            ap::Lit::Num(number) => Ok(Lit::Num(number.clone())),
+            ap::Lit::Ref(reff) => Ok(Lit::Ref(reff.clone())),
+            ap::Lit::Str(string) => Ok(Lit::Str(string.clone())),
+            ap::Lit::Symbol(symbol) => Ok(Lit::Symbol(symbol.clone())),
+            ap::Lit::Time(time) => Ok(Lit::Time(*time)),
+            ap::Lit::Uri(uri) => Ok(Lit::Uri(uri.clone())),
+            ap::Lit::YearMonth(ym) => Ok(Lit::YearMonth(ym.into())),
         }
     }
 }
 
-impl TryFrom<&Val> for LitInner {
+impl TryFrom<&Val> for Lit {
     type Error = ();
 
     fn try_from(val: &Val) -> Result<Self, Self::Error> {
@@ -1855,11 +1947,11 @@ mod tests {
     use std::convert::TryInto;
 
     fn lit_str(s: &str) -> Lit {
-        Lit::new(LitInner::Str(s.to_owned()))
+        Lit::Str(s.to_owned())
     }
 
     fn lit_num(n: f64) -> Lit {
-        Lit::new(LitInner::Num(Number::new(n, None)))
+        Lit::Num(Number::new(n, None))
     }
 
     fn idtn(tag_name: &str) -> Id {
@@ -1887,8 +1979,8 @@ mod tests {
     #[test]
     fn val_to_lit_works() {
         let val = &ap_parse(r#"{type:"literal", val:"hello"}"#).unwrap();
-        let expected = LitInner::Str("hello".to_owned());
-        let lit: LitInner = val.try_into().unwrap();
+        let expected = Lit::Str("hello".to_owned());
+        let lit: Lit = val.try_into().unwrap();
         assert_eq!(lit, expected);
     }
 
@@ -1896,8 +1988,8 @@ mod tests {
     fn val_to_lit_symbol_works() {
         let val = &ap_parse(r#"{type:"literal", val:^steam-boiler}"#).unwrap();
         let expected =
-            LitInner::Symbol(Symbol::new("^steam-boiler".to_owned()).unwrap());
-        let lit: LitInner = val.try_into().unwrap();
+            Lit::Symbol(Symbol::new("^steam-boiler".to_owned()).unwrap());
+        let lit: Lit = val.try_into().unwrap();
         assert_eq!(lit, expected);
     }
 
@@ -2289,6 +2381,85 @@ mod tests {
         assert_eq!(neg, expected);
     }
 
+    #[test]
+    fn val_to_comp_works() {
+        let val = &ap_parse(r#"{type:"compdef", params:[{name:"cells"}], body:{type:"block", exprs:[{type:"assign", lhs:{type:"var", name:"out"}, rhs:{type:"var", name:"in"}}]}, cells:{in:{is:^number, defVal:0}, out:{is:^number, ro}}}"#).unwrap();
+        let mut cells = HashMap::new();
+
+        let num = Lit::Symbol(Symbol::new("^number".into()).unwrap());
+        let zero = Lit::Num(Number::new_unitless(0.0));
+        let mut in_dict_map = HashMap::new();
+        in_dict_map.insert(tn("is"), DictVal::Expr(Expr::Lit(num.clone())));
+        in_dict_map.insert(tn("defVal"), DictVal::Expr(Expr::Lit(zero)));
+        let in_dict = Dict::new(in_dict_map);
+
+        let mut out_dict_map = HashMap::new();
+        out_dict_map.insert(tn("is"), DictVal::Expr(Expr::Lit(num)));
+        out_dict_map.insert(tn("ro"), DictVal::Marker);
+        let out_dict = Dict::new(out_dict_map);
+
+        cells.insert(tn("in"), in_dict);
+        cells.insert(tn("out"), out_dict);
+
+        let body = Block::new(vec![Expr::Assign(Assign::new(
+            tn("out"),
+            Expr::Id(Id::new(tn("in"))),
+        ))]);
+        let expected = Comp::new(cells, body);
+
+        let comp: Comp = val.try_into().unwrap();
+
+        assert_eq!(expected, comp);
+    }
+
+    #[test]
+    fn val_to_comp_with_nested_dict_works() {
+        let val = &ap_parse(r#"{type:"compdef", params:[{name:"cells"}], body:{type:"block", exprs:[{type:"var", name:"in"}]}, cells:{in:{defVal:{a:1}}}}"#).unwrap();
+        let mut cells = HashMap::new();
+
+        let mut nested_map = HashMap::new();
+        let one = Lit::Num(Number::new_unitless(1.0));
+        nested_map.insert(tn("a"), DictVal::Expr(Expr::Lit(one)));
+        let nested_dict = Dict::new(nested_map);
+
+        let mut in_dict_map = HashMap::new();
+        in_dict_map
+            .insert(tn("defVal"), DictVal::Expr(Expr::Dict(nested_dict)));
+        let in_dict = Dict::new(in_dict_map);
+
+        cells.insert(tn("in"), in_dict);
+
+        let body = Block::new(vec![Expr::Id(Id::new(tn("in")))]);
+        let expected = Comp::new(cells, body);
+
+        let comp: Comp = val.try_into().unwrap();
+
+        assert_eq!(expected, comp);
+    }
+
+    #[test]
+    fn val_to_comp_with_nested_list_works() {
+        let val = &ap_parse(r#"{type:"compdef", params:[{name:"cells"}], body:{type:"block", exprs:[{type:"var", name:"in"}]}, cells:{in:{defVal:[1]}}}"#).unwrap();
+        let mut cells = HashMap::new();
+
+        let one = Lit::Num(Number::new_unitless(1.0));
+        let nested_list = List::new(vec![Expr::Lit(one)]);
+
+        let mut in_dict_map = HashMap::new();
+        in_dict_map
+            .insert(tn("defVal"), DictVal::Expr(Expr::List(nested_list)));
+        let in_dict = Dict::new(in_dict_map);
+
+        cells.insert(tn("in"), in_dict);
+
+        let body = Block::new(vec![Expr::Id(Id::new(tn("in")))]);
+        let expected = Comp::new(cells, body);
+
+        let comp: Comp = val.try_into().unwrap();
+
+        assert_eq!(expected, comp);
+    }
+
     #[test]
     fn misc_func_works() {
         let val =
